@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997, 2005, 2006, 2007, 2009 Sergey Poznyakoff
+   Copyright (C) 1997, 2005-2007, 2009-2011, 2014-2016 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,13 +11,12 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public
-   License along with GNU cflow; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include <cflow.h>
 #include <parser.h>
+#include <ctype.h>
 
 typedef struct {
      char *name;
@@ -84,6 +83,9 @@ print_token(TOKSTK *tokptr)
      case WORD:
      case MODIFIER:
      case STRUCT:
+     case PARM_WRAPPER:
+     case QUALIFIER:
+     case OP:
 	  fprintf(stderr, "`%s'", tokptr->token);
 	  break;
      case LBRACE0:
@@ -103,21 +105,109 @@ print_token(TOKSTK *tokptr)
      case TYPEDEF:
 	  fprintf(stderr, "`typedef'");
 	  break;
-     case OP:
-	  fprintf(stderr, "OP"); /* ouch!!! */
+     case STRING:
+	  fprintf(stderr, "\"%s\"", tokptr->token);
 	  break;
      default:
 	  fprintf(stderr, "`%c'", tokptr->type);
      }
 }
 
+static char *
+token_type_str(int t)
+{
+     static char buf[80];
+     switch (t) {
+     case 0:
+	  return "EOF";
+     case WORD:
+	  return "WORD";
+     case LBRACE0:
+	  return "'{'";
+     case RBRACE0:
+	  return "'}'";
+     case IDENTIFIER:
+	  return "IDENTIFIER";
+     case EXTERN:
+	  return "EXTERN";
+     case STATIC:
+	  return "STATIC";
+     case TYPEDEF:
+	  return "TYPEDEF";
+     case STRUCT:
+	  return "STRUCT";
+     case MODIFIER:
+	  return "MODIFIER";
+     case OP:
+	  return "OP";
+     case UNION:
+	  return "UNION";
+     case ENUM:
+	  return "ENUM";
+     case LBRACE:
+	  return "' {'";
+     case RBRACE:
+	  return "' }'";
+     case MEMBER_OF:
+	  return "MEMBER_OF";
+     case TYPE:
+	  return "TYPE";
+     case STRING:
+	  return "STRING";
+     case PARM_WRAPPER:
+	  return "PARM_WRAPPER";
+     case QUALIFIER:
+	  return "QUALIFIER";
+     }	  
+     if (isprint(t))
+	  snprintf(buf, sizeof(buf), "'%c'(%d)", t, t);
+     else
+	  snprintf(buf, sizeof(buf), "%d", t);
+     return buf;
+}
+
 static void
-file_error(char *msg, int near)
+dbgtok(TOKSTK *t, int delim)
+{
+     if (delim)
+	  fputc(delim, stderr);
+     fprintf(stderr, "{ %s ", token_type_str(t->type));
+     if (t->type)
+	  fprintf(stderr, ", %s, %d ", t->token ? t->token : "NULL", t->line);
+     fputc('}', stderr);
+}
+
+static void
+debugtoken(TOKSTK *t, char *fmt, ...)
+{
+     if (debug > 1) {
+	  va_list ap;
+	  int i;
+	  
+	  if (fmt) {
+	       va_start(ap, fmt);
+	       vfprintf(stderr, fmt, ap);
+	       va_end(ap);
+	       fprintf(stderr, ": ");
+	  }
+	  if (t) {
+	       dbgtok(t, 0);
+	       fprintf(stderr, "; ");
+	  }
+	  fprintf(stderr, "%d: {", curs);
+	  for (i = curs; i < tos; i++)
+	       dbgtok(token_stack + i, i == curs ? 0 : ',');
+	  fprintf(stderr, "}\n");
+     }
+}
+
+static void
+file_error(char *msg, TOKSTK *tokptr)
 {
     fprintf(stderr, "%s:%d: %s", filename, tok.line, msg);
-    if (near) {
+    if (tokptr) {
 	fprintf(stderr, _(" near "));
-	print_token(&tok);
+	print_token(tokptr);
     }
     fprintf(stderr, "\n");
 }
@@ -126,14 +216,45 @@ void
 mark(Stackpos pos)
 {
      pos[0] = curs;
+     if (debug > 1)
+	  fprintf(stderr, "marking stack at %d\n", curs);
 }
 
 void
 restore(Stackpos pos)
 {
      curs = pos[0];
-     if (curs)
+     if (curs) {
 	  tok = token_stack[curs-1];
+	  debugtoken(&tok, "restored stack");
+     }
+}
+
+void
+tokdel(int beg, int end)
+{
+     if (end >= beg) {
+	  if (end < tos)
+	       memmove(token_stack + beg, token_stack + end + 1,
+		       (end - beg + 1) * sizeof(token_stack[0]));
+	  tos -= (end - beg + 1);
+     }
+}
+
+void
+tokins(int pos, int type, int line, char *token)
+{
+     if (++tos == token_stack_length) {
+	  token_stack_length += token_stack_increase;
+	  token_stack = xrealloc(token_stack,
+				 token_stack_length*sizeof(*token_stack));
+     }
+     memmove(token_stack + pos + 1, token_stack + pos,
+	     (tos - pos - 1) * sizeof(token_stack[0]));
+     token_stack[pos].type = type;
+     token_stack[pos].token = token;
+     token_stack[pos].line = line;
+     debugtoken(&token_stack[pos], "insert at %d", pos);
 }
 
 void
@@ -154,9 +275,10 @@ cleanup_stack()
 {
      int delta = tos - curs;
 
-     if (delta) 
+     if (delta > 0) 
 	  memmove(token_stack, token_stack+curs, delta*sizeof(token_stack[0]));
-
+     else if (delta < 0) /* Invalid input */
+	  delta = 0;
      tos = delta;
      curs = 0;
 }
@@ -175,23 +297,27 @@ nexttoken()
      if (curs == tos) {
 	  type = get_token();
 	  tokpush(type, line_num, yylval.str);
+	  yylval.str = NULL;
      }
      tok = token_stack[curs];
      curs++;
+     debugtoken(&tok, "next token");
      return tok.type;
 }
 
 int
 putback()
 {
-     if (curs == 0)
-	  error(10, 0, _("INTERNAL ERROR: cannot return token to stream"));
+     if (curs == 0) {
+	  error(0, 0, _("INTERNAL ERROR: cannot return token to stream"));
+	  abort();
+     }
      curs--;
      if (curs > 0) {
-	  tok.type = token_stack[curs-1].type;
-	  tok.token = token_stack[curs-1].token;
+	  tok = token_stack[curs-1];
      } else
 	  tok.type = 0;
+     debugtoken(&tok, "putback");
      return tok.type;
 }
 
@@ -214,6 +340,7 @@ save_token(TOKSTK *tokptr)
      case STRUCT:
      case PARM_WRAPPER:
      case WORD:
+     case QUALIFIER:
 	  if (need_space) 
 	       obstack_1grow(&text_stk, ' ');
 	  len = strlen(tokptr->token);
@@ -233,13 +360,33 @@ save_token(TOKSTK *tokptr)
      case EXTERN: /* storage class specifiers are already taken care of */
      case STATIC:
 	  break;
+     case ',':
+	  obstack_1grow(&text_stk, ',');
+	  need_space = 1;
+	  break;
      case '(':
 	  if (need_space) 
 	       obstack_1grow(&text_stk, ' ');
-	  /*FALLTHRU*/
-     default:
 	  obstack_1grow(&text_stk, tokptr->type);
 	  need_space = 0;
+	  break;
+     case ')':
+	  obstack_1grow(&text_stk, tokptr->type);
+	  need_space = 1;
+	  break;
+     case '[':
+     case ']':
+	  obstack_1grow(&text_stk, tokptr->type);
+	  need_space = 0;
+	  break;
+     case OP:
+	  obstack_1grow(&text_stk, ' ');
+	  obstack_grow(&text_stk, tokptr->token, strlen(tokptr->token));
+	  need_space = 1;
+	  break;
+     default:
+	  if (verbose)
+	       file_error(_("unrecognized definition"), tokptr);
      }
 }
 
@@ -304,6 +451,33 @@ skip_to(int c)
      }
 }
 
+int
+skip_balanced(int open_tok, int close_tok, int level)
+{
+     if (level == 0) {
+	  if (nexttoken() != open_tok) {
+	       return 1;
+	  }
+	  level++;
+     }
+     while (nexttoken()) {
+	  if (tok.type == LBRACE0 && open_tok == '{')
+	       tok.type = '{';
+	  else if (tok.type == RBRACE0 && close_tok == '}')
+	       tok.type = '}';
+	  
+	  if (tok.type == open_tok) 
+	       level++;
+	  else if (tok.type == close_tok) {
+	       if (--level == 0) {
+		    nexttoken();
+		    return 0;
+	       }
+	  } 
+     }
+     return -1;
+}
+
 int
 yyparse()
 {
@@ -317,6 +491,8 @@ yyparse()
 	  switch (tok.type) {
 	  case 0:
 	       return 0;
+	  case QUALIFIER:
+	       continue;
 	  case TYPEDEF:
 	       parse_typedef();
 	       break;
@@ -334,7 +510,7 @@ yyparse()
 	  }
 	  cleanup_stack();
      }
-     /*NOTREACHED*/
+     return 0;
 }
 
 static int
@@ -344,17 +520,32 @@ is_function()
      int res = 0;
 
      mark(sp);
-/*    if (tok.type == STRUCT)
-	nexttoken();*/
-     while (tok.type == TYPE ||
-	    tok.type == IDENTIFIER ||
-	    tok.type == MODIFIER ||
-	    tok.type == STATIC ||
-	    tok.type == EXTERN)
-	  nexttoken();
-     
-     if (tok.type == '(') 
-	  res = nexttoken() != MODIFIER;
+     while (1) {
+	  switch (tok.type) {
+	  case QUALIFIER:
+	  case TYPE:
+	  case IDENTIFIER:
+	  case MODIFIER:
+	  case STATIC:
+	  case EXTERN:
+	  case STRUCT:
+	  case UNION:
+	  case ENUM:
+	       nexttoken();
+	       continue;
+	  case PARM_WRAPPER:
+	       if (skip_balanced('(', ')', 0) == -1)
+		    file_error(_("unexpected end of file in declaration"),
+			       NULL);
+	       continue;
+	  case '(':
+	       res = nexttoken() != MODIFIER;
+	       break;
+	  default:
+	       break;
+	  }
+	  break;
+     }
      
      restore(sp);
      return res;
@@ -395,7 +586,8 @@ expression()
 	       break;
 	  case 0:
 	       if (verbose)
-		    file_error(_("unexpected end of file in expression"), 0);
+		    file_error(_("unexpected end of file in expression"),
+			       NULL);
 	       return;
 	    
 	  case IDENTIFIER:
@@ -450,7 +642,7 @@ parse_function_declaration(Ident *ident, int parm)
 	       nexttoken();
 	  else {
 	       if (verbose) 
-		    file_error(_("expected `;'"), 1);
+		    file_error(_("expected `;'"), &tok);
 	       error_recovery = 1;
 	  }
 	  goto restart;
@@ -462,12 +654,14 @@ parse_function_declaration(Ident *ident, int parm)
      case LBRACE:
 	  if (ident->name) {
 	       caller = lookup(ident->name);
+	       if (caller && caller->storage == AutoStorage)
+		    caller = NULL;
 	       func_body();
 	  }
 	  break;
      case 0:
 	  if (verbose)
-	       file_error(_("unexpected end of file in declaration"), 0);
+	       file_error(_("unexpected end of file in declaration"), NULL);
      }
 }
 
@@ -480,27 +674,23 @@ fake_struct(Ident *ident)
      ident->type_end = -1;
      if (tok.type == STRUCT) {
 	  if (nexttoken() == IDENTIFIER) {
-	       ident->type_end = tos;
+	       ident->type_end = curs;
 	  }
 	  putback();
 	  skip_struct();
 	  if (tok.type == IDENTIFIER || tok.type == MODIFIER) {
-	       TOKSTK hold = tok;
+	       int pos = curs-1;
 	       restore(sp);
 	       if (ident->type_end == -1) {
 		    /* there was no tag. Insert { ... } */
-		    tos = curs;
-		    token_stack[curs].type = IDENTIFIER;
-		    token_stack[curs].token = "{ ... }";
-		    tos++;
-	       } else {
-		    tos = curs + 1;
+		    tokdel(curs, pos - 1);
+		    tokins(curs, IDENTIFIER, tok.line, "{ ... }");
+		    debugtoken(&tok, "modified stack");
 	       }
-	       tokpush(hold.type, hold.line, hold.token);
-	  } else {
-	       if (tok.type != ';')
-		    file_error(_("missing `;' after struct declaration"), 0);
-	  }
+	  } else if (tok.type == '(')
+	       return 0;
+	  else if (tok.type != ';')
+	       file_error(_("missing `;' after struct declaration"), &tok);
 	  return 1;
      }
      return 0;
@@ -513,27 +703,23 @@ parse_variable_declaration(Ident *ident, int parm)
      
      mark(sp);
      ident->type_end = -1;
-     if (tok.type == STRUCT) {
+     if (tok.type == STRUCT || tok.type == UNION) {
 	  if (nexttoken() == IDENTIFIER) {
-	       ident->type_end = tos;
+	       ident->type_end = curs;
 	  }
 	  putback();
 	  skip_struct();
-	  while (tok.type == MODIFIER)
+	  while (tok.type == MODIFIER || tok.type == QUALIFIER)
 	       nexttoken();
 	  if (tok.type == IDENTIFIER) {
-	       TOKSTK hold = tok;
+	       int pos = curs-1;
 	       restore(sp);
 	       if (ident->type_end == -1) {
 		    /* there was no tag. Insert { ... } */
-		    tos = curs;
-		    token_stack[curs].type = IDENTIFIER;
-		    token_stack[curs].token = "{ ... }";
-		    tos++;
-	       } else {
-		    tos = curs + 1;
+		    tokdel(curs, pos - 1);
+		    tokins(curs, IDENTIFIER, tok.line, "{ ... }");
+		    debugtoken(&tok, "modified stack");
 	       }
-	       tokpush(hold.type, hold.line, hold.token);
 	  } else {
 	       if (tok.type == ';')
 		    return;
@@ -551,7 +737,7 @@ parse_variable_declaration(Ident *ident, int parm)
 	  /*FALLTHROUGH*/
      default:
 	  if (verbose) 
-	       file_error(_("expected `;'"), 1);
+	       file_error(_("expected `;'"), &tok);
 	  /* FIXME: should putback() here */
 	  /* FALLTHRU */
      case ';':
@@ -576,7 +762,7 @@ parse_variable_declaration(Ident *ident, int parm)
 	  break;
      case 0:
 	  if (verbose)
-	       file_error(_("unexpected end of file in declaration"), 0);
+	       file_error(_("unexpected end of file in declaration"), NULL);
      }
 }
 
@@ -598,7 +784,8 @@ initializer_list()
 	       }
 	       break;
 	  case 0:
-	       file_error(_("unexpected end of file in initializer list"), 0);
+	       file_error(_("unexpected end of file in initializer list"),
+			  NULL);
 	       return;
 	  case ',':
 	       break;
@@ -620,29 +807,21 @@ parse_knr_dcl(Ident *ident)
 void
 skip_struct()
 {
-     int lev = 0;
-     
      if (nexttoken() == IDENTIFIER) {
 	  nexttoken();
      } else if (tok.type == ';')
 	  return;
      
      if (tok.type == LBRACE || tok.type == LBRACE0) {
-	  do {
-	       switch (tok.type) {
-	       case 0:
-		    file_error(_("unexpected end of file in struct"), 0);
-		    return;
-	       case LBRACE:
-	       case LBRACE0:
-		    lev++;
-		    break;
-	       case RBRACE:
-	       case RBRACE0:
-		    lev--;
-	       }
-	       nexttoken();
-	  } while (lev);
+	  if (skip_balanced('{', '}', 1) == -1) {
+	       file_error(_("unexpected end of file in struct"), NULL);
+	       return;
+	  }
+     }
+
+     while (tok.type == PARM_WRAPPER) {
+	  if (skip_balanced('(', ')', 0) == -1)
+	      file_error(_("unexpected end of file in struct"), NULL);
      }
 }
 
@@ -683,22 +862,27 @@ parse_dcl(Ident *ident, int maybe_knr)
 int
 dcl(Ident *idptr)
 {
-     int type;
-
      while (nexttoken() != 0 && tok.type != '(') {
 	  if (tok.type == MODIFIER) {
 	       if (idptr && idptr->type_end == -1)
 		    idptr->type_end = curs-1;
+	  } else if (tok.type == PARM_WRAPPER) {
+	       if (skip_balanced('(', ')', 0) == -1) {
+		    file_error(_("unexpected end of file in function declaration"),
+			       NULL);
+		    return 1;
+	       }
 	  } else if (tok.type == IDENTIFIER) {
+	       int type;
+	       
 	       while (tok.type == IDENTIFIER)
 		    nexttoken();
 	       type = tok.type;
 	       putback();
-	       if (type == TYPE)
-		    continue;
-	       else if (type != MODIFIER) 
+	       if (!(type == TYPE ||
+		     type == MODIFIER || type == QUALIFIER))
 		    break;
-	  } else if (tok.type == ')') {
+	  } else if (tok.type == ')' || tok.type == ';') {
 	       return 1;
 	  }
      }
@@ -716,7 +900,7 @@ dirdcl(Ident *idptr)
      if (tok.type == '(') {
 	  dcl(idptr);
 	  if (tok.type != ')' && verbose) {
-	       file_error(_("expected `)'"), 1);
+	       file_error(_("expected `)'"), &tok);
 	       return 1;
 	  }
      } else if (tok.type == IDENTIFIER) {
@@ -739,7 +923,7 @@ dirdcl(Ident *idptr)
 	  else {
 	       maybe_parm_list(parm_ptr);
 	       if (tok.type != ')' && verbose) {
-		    file_error(_("expected `)'"), 1);
+		    file_error(_("expected `)'"), &tok);
 		    return 1;
 	       }
 	  }
@@ -747,26 +931,12 @@ dirdcl(Ident *idptr)
      if (wrapper)
 	  nexttoken(); /* read ')' */
 
-     if (tok.type == PARM_WRAPPER) {
-	  if (nexttoken() == '(') {
-	       int level = 0;
-	       while (nexttoken()) {
-		    if (tok.type == 0) {
-			 file_error(_("unexpected end of file in function declaration"),
-				    0);
-			 return 1;
-		    } else if (tok.type == '(') 
-			 level++;
-		    else if (tok.type == ')') {
-			 if (level-- == 0) {
-			      nexttoken();
-			      break;
-			 }
-		    } 
-	       }
-	  } else
-	       putback();
+     while (tok.type == PARM_WRAPPER) {
+	  if (skip_balanced('(', ')', 0) == -1)
+	       file_error(_("unexpected end of file in function declaration"),
+			  NULL);
      }
+     
      return 0;
 }
 
@@ -812,10 +982,11 @@ maybe_parm_list(int *parm_cnt_return)
 	       return;
 	  case ',':
 	       break;
+	  case QUALIFIER:
 	  case IDENTIFIER:
+	  case MODIFIER: /* unsigned * */
 	  case STRUCT:
 	  case UNION:
-	  case ENUM:
 	  case TYPE:
 	       parmcnt++;
 	       ident.storage = AutoStorage;
@@ -824,7 +995,8 @@ maybe_parm_list(int *parm_cnt_return)
 	       break;
 	  default:
 	       if (verbose)
-		    file_error(_("unexpected token in parameter list"), 1);
+		    file_error(_("unexpected token in parameter list"),
+			       &tok);
 	       level = 0;
 	       do {
 		    if (tok.type == '(') 
@@ -839,7 +1011,7 @@ maybe_parm_list(int *parm_cnt_return)
 	  }
      }
      if (verbose)
-	  file_error(_("unexpected end of file in parameter list"), 0);
+	  file_error(_("unexpected end of file in parameter list"), NULL);
 }
 
 void
@@ -877,7 +1049,7 @@ func_body()
 	  case RBRACE0:
 	       if (use_indentation) {
 		    if (verbose && level != 1)
-			 file_error(_("forced function body close"), 0);
+			 file_error(_("forced function body close"), NULL);
 		    for ( ; level; level--) {
 			 delete_autos(level);
 		    }
@@ -891,10 +1063,13 @@ func_body()
 	       break;
 	  case 0:
 	       if (verbose)
-		    file_error(_("unexpected end of file in function body"), 0);
+		    file_error(_("unexpected end of file in function body"),
+			       NULL);
+	       caller = NULL;
 	       return;
 	  }
      }
+     caller = NULL;
 }
 
 int
@@ -969,7 +1144,8 @@ declare(Ident *ident, int maybe_knr)
 
      if ((ident->parmcnt >= 0
 	  && (!maybe_knr || get_knr_args(ident) == 0)
-	  && !(tok.type == LBRACE || tok.type == LBRACE0 || tok.type == TYPE))
+	  && !(tok.type == LBRACE || tok.type == LBRACE0 || tok.type == TYPE
+	       || tok.type == PARM_WRAPPER))
 	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)) {
 	  undo_save_stack();
 	  /* add_external()?? */
@@ -982,9 +1158,14 @@ declare(Ident *ident, int maybe_knr)
 	      && (sp->storage != StaticStorage || level > 0)) {
 	       sp = install_ident(ident->name, ident->storage);
 	  } else {
-	       error_at_line(0, 0, filename, ident->line, 
-			     _("%s/%d redefined"),
-			     ident->name, sp->arity);
+	       if (sp->arity >= 0)
+		    error_at_line(0, 0, filename, ident->line, 
+				  _("%s/%d redefined"),
+				  ident->name, sp->arity);
+	       else
+		    error_at_line(0, 0, filename, ident->line, 
+				  _("%s redefined"),
+				  ident->name);
 	       error_at_line(0, 0, sp->source, sp->def_line,
 			     _("this is the place of previous definition"));
 	  }
@@ -1000,7 +1181,7 @@ declare(Ident *ident, int maybe_knr)
      sp->def_line = ident->line;
      sp->level = level;
      if (debug)
-	  printf(_("%s:%d: %s/%d defined to %s\n"),
+	  fprintf(stderr, _("%s:%d: %s/%d defined to %s\n"),
 		 filename,
 		 line_num,
 		 ident->name, ident->parmcnt,
@@ -1018,17 +1199,15 @@ declare_type(Ident *ident)
 	  if (sp->type == SymToken && sp->token_type == TYPE)
 	       break;
      if (!sp)
-	  sp = install(ident->name);
+	  sp = install(ident->name, INSTALL_UNIT_LOCAL);
      sp->type = SymToken;
      sp->token_type = TYPE;
      sp->source = filename;
      sp->def_line = ident->line;
      sp->ref_line = NULL;
      if (debug)
-	  printf(_("%s:%d: type %s\n"),
-		 filename,
-		 line_num,
-		 ident->name);
+	  fprintf(stderr, _("%s:%d: type %s\n"), filename, line_num,
+		  ident->name);
 }
 
 Symbol *
