@@ -1,5 +1,5 @@
 /*  This file is part of GNU cflow
- *  Copyright (C) 1997, 2007 Sergey Poznyakoff
+ *  Copyright (C) 1997, 2007, 2010, 2014-2016 Sergey Poznyakoff
  *  
  *  GNU cflow is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -12,9 +12,7 @@
  *  GNU General Public License for more details.
  *  
  *  You should have received a copy of the GNU General Public License
- *  along with GNU cflow; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *  MA 02110-1301 USA
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 /* tokens */
 #define WORD 257
@@ -35,6 +33,7 @@
 #define TYPE 270
 #define STRING 271
 #define PARM_WRAPPER 272
+#define QUALIFIER 273
 
 typedef struct {
     char *str;
